/*

  Le Pin - A multiplayer game for saturn
  latest
  
  Copyright (C) 2002 Vivien Guillet
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include    "sl_def.h"
#include    "sgl.h"
#include    "sddrvs.dat"    // sound driver
#include    "scrltxt.h"     // scrolling display

typedef enum _gss {gss_on,gss_off} gss;
static gss GLOBAL_SOUND_STATE = gss_off;

/********************************************************
 * SON
 ********************************************************/

extern char s_16[];
extern Uint32 s_16_size;
extern PCM s_16_dat ;

void init_sound(){
    char sound_map[] = {0xff , 0xff} ;
    slInitSound(sddrvstsk , sizeof(sddrvstsk) , (Uint8 *)sound_map , sizeof(sound_map)) ;
}

void playasound(){
  if (GLOBAL_SOUND_STATE == gss_off) {
    s_16_dat.pitch = (Uint16)(slMulFX(slRandom(),(FIXED)0x0000));
    slPCMOn(&s_16_dat , s_16 , s_16_size) ;
  }
}


/*********************************************************
 * DEFINITIONS SGL
 *********************************************************/
 
/*
static Uint16 pad_asign[] = {
	PER_DGT_KU,
	PER_DGT_KD,
	PER_DGT_KR,
	PER_DGT_KL,
	PER_DGT_TA,
	PER_DGT_TB,
	PER_DGT_TC,
	PER_DGT_ST,
	PER_DGT_TX,
	PER_DGT_TY,
	PER_DGT_TZ,
	PER_DGT_TR,
	PER_DGT_TL,
};
*/


/*********************************************************
 * DEFINES
 *********************************************************/
#define TRUE (0==0)
#define FALSE (0!=0)

// for pos[-1,1] -> screen
// acc to  slInitSystem(TV_320x224,NULL,1);

#define S_LIM_L toFIXED(-1.0)
#define S_LIM_R toFIXED(1.0)
#define S_LIM_T toFIXED(-0.8)
#define S_LIM_B toFIXED(0.8)
#define L_XCOEF toFIXED(200.0)
#define L_YCOEF toFIXED(160.0)

#define MAX_SATURN_PAD 15

// ! These defines represents the last array element, not the size  
#define MAX_PLAYER 9     
#define MAX_PLANE  4    
#define MAX_BOMB  20   
#define MAX_PLAYER_ENERGY 100

#define LIMBO_TIME 100   

#define MAX_COLL (MAX_PLAYER+1) // for collision computation

#define COSPISUR4 toFIXED(0.707)


/*********************************************************
 * STRUCTURES
 *********************************************************/

typedef struct meca_vector {
    FIXED x;
    FIXED y;
    FIXED dx;
    FIXED dy;   
} MECA_VECTOR ;

// player states
#define PL_STATE_PLAY 1
#define PL_STATE_DEAD 0
#define PL_STATE_LIMBO -1

typedef struct player {
   // sys
   int state;		// PL_STATE (if DEAD, the struct state is undef)
   int r_limbo_time;	// when state = PL_STATE_LIMBO, remaining time before death
   int plane_num;	
   int pad_num;		// pad number in smpc
   Uint16 prev_pad_data;// used to detect fronts
   // util
   MECA_VECTOR meca;	
   int energy;		// <= MAX_ENERGY
   int bonus;		// TODO: unused for the moment

} PLAYER;

// bombs state
#define BB_STATE_NIU   1  // not in use
#define BB_STATE_CDOWN 2  // count down
#define BB_STATE_EXPL  3  // explose

typedef struct plane {
  PDATA* ppdata; 		// sgl au loin !
  FIXED speedLimit;
  FIXED accel;
  FIXED mass;	// [0..1]
} PLANE;

typedef struct bomb {
  PDATA* ppdata;
  int    type;
  int    state;
  int    cdown_time;	// temps avant l'explosion
  int    expl_time;	// temps d'explosion
  int    r_cdown_time;	// temps restant avant l'explosion
  int    r_expl_time;	// temps d'explosion restant 
  FIXED  expl_radius;	// rayon d'explosion max
  FIXED  c_expl_radius;	// rayon d'explosion 
  
  MECA_VECTOR meca;
  
} BOMB;

typedef struct coll {
  int  isUsed;
  int  player[MAX_PLAYER+1];
  int  bomb[MAX_BOMB+1];
  int  wall[4];
}COLL ;

/*********************************************************
 * GLOB
 *********************************************************/

PLAYER Player[MAX_PLAYER+1];	// tableau contenant les joueurs
PLANE  Plane[MAX_PLANE+1];	// tableau contenant les avions
BOMB   Bomb[MAX_BOMB+1];	// tableau contenant les bombes
COLL   Coll[MAX_COLL+1];        // tableau des collisions

/*********************************************************
 * BOMB MANAGEMENT
 ********************************************************/
// gfx

POINT point_BOMB1[]= {
  POStoFIXED(-5.0,-5.0, 0.0),  //0 
  POStoFIXED( 5.0,-5.0, 0.0),  //1
  POStoFIXED( 0.0,-1.0, 0.0),  //2
  POStoFIXED(-5.0,-5.0, 0.0),  //3=0
  
  POStoFIXED(-5.0, 5.0, 0.0),  //4
  POStoFIXED( 5.0, 5.0, 0.0),  //5
  POStoFIXED( 0.0, 1.0, 0.0),  //6
  POStoFIXED(-5.0, 5.0, 0.0),  //7
};

POLYGON polygon_BOMB1[]= {
  NORMAL(0.0,1.0,0.0),
  VERTICES(0,1,2,3),
  NORMAL(0.0,1.0,0.0),
  VERTICES(4,5,6,7),
};

ATTR attribute_BOMB1[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(31,31,31),No_Gouraud,MESHoff,sprPolygon,No_Option),  
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(31,0,0),No_Gouraud,MESHoff,sprPolygon,No_Option),  
};


PDATA PD_BOMB1 = {
  point_BOMB1,sizeof(point_BOMB1)/sizeof(POINT),
  polygon_BOMB1,sizeof(polygon_BOMB1)/sizeof(POLYGON),
  attribute_BOMB1
};

PDATA PD_BOMB2 = {
  point_BOMB1,sizeof(point_BOMB1)/sizeof(POINT),
  polygon_BOMB1,sizeof(polygon_BOMB1)/sizeof(POLYGON),
  attribute_BOMB1
};

// init
void buildBombArray()  {

  int i;
  for (i=0 ; i<=MAX_BOMB ; i++) { 
    Bomb[i].state = BB_STATE_NIU;
    Bomb[i].meca.x = toFIXED(0.0);
    Bomb[i].meca.y = toFIXED(0.0);
    Bomb[i].meca.dx = toFIXED(0.0);
    Bomb[i].meca.dy = toFIXED(0.0);
    Bomb[i].type = 0;
    Bomb[i].cdown_time = 200;
    Bomb[i].expl_time = 50;
    Bomb[i].c_expl_radius = toFIXED(0.0);     
    Bomb[i].expl_radius = toFIXED(0.5);
    Bomb[i].ppdata= &(PD_BOMB1);
  }
  

  //  Bomb[1].ppdata=&(PD_BOMB2);  
  
}


void armBomb(int bombnum){
  Bomb[bombnum].r_cdown_time = Bomb[bombnum].cdown_time;
  Bomb[bombnum].r_expl_time = Bomb[bombnum].expl_time ;
  Bomb[bombnum].c_expl_radius = toFIXED(0.0) ;       
}

// can drop a bomb if one is remaining 
int askLaunchBomb(FIXED x,FIXED y) {
  int i,bombnum;
  
  bombnum = -1;
  for (i = 0; i <= MAX_BOMB; i++) {
    if (Bomb[i].state == BB_STATE_NIU) {
      bombnum = i;
    }
  }
  
  if (bombnum != -1) {
    Bomb[bombnum].state = BB_STATE_CDOWN;
    Bomb[bombnum].meca.x = x;


    Bomb[bombnum].meca.y = y;
    armBomb(bombnum);
  } 
  return bombnum;
}

/*********************************************************
 * PLANE MANAGEMENT
 *********************************************************/

// TODO : different gfx

POINT point_PLANE1[]= {
  POStoFIXED(-5.0,-5.0,0.0),
  POStoFIXED( 5.0,-5.0,0.0),
  POStoFIXED( 5.0, 5.0,0.0),
  POStoFIXED(-5.0, 5.0,0.0),
};

POLYGON polygon_PLANE1[]= {
  NORMAL(0.0,1.0,0.0),
  VERTICES(0,1,2,3),
};

ATTR attribute_PLANE1[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(31,0,0),No_Gouraud,MESHoff,sprPolygon,No_Option),  
};
ATTR attribute_PLANE2[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(0,31,0),No_Gouraud,MESHoff,sprPolygon,No_Option),  
};

ATTR attribute_PLANE3[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(0,0,31),No_Gouraud,MESHoff,sprPolygon,No_Option),  
};

ATTR attribute_PLANE4[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(31,31,0),No_Gouraud,MESHoff,sprPolygon,No_Option),  
};

ATTR attribute_PLANE5[]= {
  ATTRIBUTE(Dual_Plane,SORT_CEN,No_Texture,C_RGB(31,0,31),No_Gouraud,MESHoff,sprPolygon,No_Option),  
};


PDATA PD_PLANE1 = {
  point_PLANE1,sizeof(point_PLANE1)/sizeof(POINT),
  polygon_PLANE1,sizeof(polygon_PLANE1)/sizeof(POLYGON),
  attribute_PLANE1
};

PDATA PD_PLANE2 = {
  point_PLANE1,sizeof(point_PLANE1)/sizeof(POINT),
  polygon_PLANE1,sizeof(polygon_PLANE1)/sizeof(POLYGON),
  attribute_PLANE2
};

PDATA PD_PLANE3 = {
  point_PLANE1,sizeof(point_PLANE1)/sizeof(POINT),
  polygon_PLANE1,sizeof(polygon_PLANE1)/sizeof(POLYGON),
  attribute_PLANE3
};

PDATA PD_PLANE4 = {
  point_PLANE1,sizeof(point_PLANE1)/sizeof(POINT),
  polygon_PLANE1,sizeof(polygon_PLANE1)/sizeof(POLYGON),
  attribute_PLANE4
};

PDATA PD_PLANE5 = {
  point_PLANE1,sizeof(point_PLANE1)/sizeof(POINT),
  polygon_PLANE1,sizeof(polygon_PLANE1)/sizeof(POLYGON),
  attribute_PLANE5
};



void buildPlaneArray()  {

  int i;
  
  for (i=0 ; i<=MAX_PLANE ; i++) { //TODO : caracteristiques.
	Plane[i].speedLimit = slMulFX(toFIXED(0.05),slRandom());
  	Plane[i].accel = slMulFX(toFIXED(0.005),slRandom());
  	Plane[i].mass = toFIXED(slRandom());     
  }

  Plane[0].ppdata=&(PD_PLANE1);
  Plane[1].ppdata=&(PD_PLANE2);
  Plane[2].ppdata=&(PD_PLANE3);
  Plane[3].ppdata=&(PD_PLANE4);
  Plane[4].ppdata=&(PD_PLANE5);
  for (i=0 ; i<=MAX_PLANE ; i++) { //quoi ?
    //    Bomb[i].ppdata->attbl->colno = C_RGB(31,31,31);
  }

}


/*********************************************************
 * PAD MANAGEMENT FUNCTIONS
 *********************************************************/
/*
Detection des fronts montants (0->1) :                                       
|0|1|1|0|0|1||1| ... previous            _p_
|1|0|0|1|1|0||0| ... !previous           0 1
|0|1|0|0|1|0||1| ... next           n| 0 0 1
|0|0|0|0|1|0||0| ... upfront         | 1 1 0

*/
// 1->0
Uint16 pad_GetDownFront (Uint16 prev, Uint16 cur) {
  Uint16 downfront;
  downfront = (~((~prev) & cur));  
  return downfront;
}


// 0->1
Uint16 pad_GetUpFront (Uint16 prev, Uint16 cur) {
   Uint16 upfront = 0;
   upfront = pad_GetDownFront(cur,prev);
   return upfront;
}

/*********************************************************
 * POSITION MANAGEMENT FUNCTIONS
 *********************************************************/
 
FIXED logicalXToScreenX (FIXED x) { 
	return(slMulFX(L_XCOEF,x));
}
FIXED logicalYToScreenY (FIXED y) { 
	return(slMulFX(L_YCOEF,y));
}

FIXED getDistance(FIXED x1, FIXED y1, FIXED x2,FIXED y2) {
  FIXED distance;
  FIXED dx,dy;

  dx = (x1 - x2);
  dy = (y1 - y2);
  distance =  slSquartFX(slMulFX(dx,dx)+slMulFX(dy,dy));

  return distance;

}

FIXED getSpeedFromSpeedXY(FIXED dx,FIXED dy){
  return getDistance(toFIXED(0.0),toFIXED(0.0),dx,dy);
}

void applySpeed(FIXED* px,FIXED dx) { 
	(*px) += (dx);
}

/*
 * pdx , pdy :pointeurs sur les vecteurs vitesse
 * accel acceleration dispo
 * padx, pady : -1, 0 ou 1 en fonction de ce qui est demand dans le pad
 */

//  vieux : utilis pour que l'acceleration diminue a l'approche de la vitesse max.
//  axeaccelX = slDivFX(speedlimit,slMulFX((speedlimit-(*pdx),axeaccel)));    

void applyPadAccel(FIXED* pdx,FIXED* pdy,FIXED accel, FIXED speedlimit,int padx,int pady) {

  FIXED axeaccel = accel ;
  FIXED oppSpeedlimit;

  oppSpeedlimit = slMulFX(toFIXED(-1.0),speedlimit);
  
  // on n'accelere pas dans la direction ou l'on est deja trop rapide. 
  if (((*pdx)>speedlimit)    && (padx>0)) padx = 0;
  if (((*pdx)<oppSpeedlimit) && (padx<0)) padx = 0; 
  if (((*pdy)>speedlimit)    && (pady>0)) pady = 0;
  if (((*pdy)<oppSpeedlimit) && (pady<0)) pady = 0; 

  // pour viter que l'on accelere deux fois en allant en diagonale
  if ((padx !=0) && (pady !=0)){
    axeaccel = slMulFX(accel,COSPISUR4);
  } else {
      // si l'on va dans une seule direction, on tend vers celle la.
    if (padx !=0) {
      (*pdy) = slMulFX(toFIXED(0.95),(*pdy))   ;
    }
    if (pady !=0) {
      (*pdx) = slMulFX(toFIXED(0.95),(*pdx))   ;
    }
  }

  // on applique l'acceleration calcule.
  if (padx>0) (*pdx) += axeaccel;
  if (padx<0) (*pdx) -= axeaccel;
  if (pady>0) (*pdy) += axeaccel;
  if (pady<0) (*pdy) -= axeaccel;

}

/*********************************************************
 * VOTE
 *********************************************************/
/*
typedef struct vote {
  enum vote_type {vt_boudaries,vt_bombspeed} typez;
  int       r_time;
  int       total_time;
  int       voters[MAX_PLAYER+1];       
};

initiateVote(int player){
}
*/
/*********************************************************
 * PLAYER MANAGEMENT FUNCTIONS
 *********************************************************/
/*
Le tableau des joueurs de taille MAX_PLAYER contient les diffrents
Joueurs. 
Un joueur est ajout quand 
- START est appuy, et le PLAYER.state est DEAD
Un joueur est supprim quand
- START est appy, et le PLAYER.sate est PLAY
- il meurt, etc.

L'tat state doit tre consult avant tout acces a PLAYER[n]. Si il 
est a DEAD, les autres champs sont indetermins.
*/

// renvoie un avion non attribu ou -1 si y'en a pu.

int getAFreePlane() {
  int i,j;
  int presents;
  int freeplanenum = -1;
  for (i=0;i<=MAX_PLANE;i++) {	// brutal: on prend le dernier qu'on trouve...
  	presents = 0;
	for (j=0;j<=MAX_PLAYER;j++) {
	  	if (Player[j].state == PL_STATE_PLAY) {
			presents += (Player[j].plane_num == i)?1:0;
		}
  	}
  	if (presents == 0) {
  	  freeplanenum = i; 	  
  	}
  }
  return freeplanenum;
}

void wakeUpPlayer(int playernum, int padnumber) {
  int ticket;
  ticket = getAFreePlane();
  if (ticket == -1) {
  	Player[playernum].state = PL_STATE_DEAD; // pas d'avion, pas de vie.
  } else {
    Player[playernum].plane_num = ticket;
    Player[playernum].state = PL_STATE_PLAY;
    Player[playernum].pad_num = padnumber;  	
    Player[playernum].prev_pad_data = 0;//Smpc_Peripheral[padnumber].data;
    Player[playernum].energy = MAX_PLAYER_ENERGY;
    Player[playernum].meca.x = slMulFX(slRandom(),toFIXED(2))-toFIXED(1.0);
    Player[playernum].meca.y = slMulFX(slRandom(),toFIXED(2))-toFIXED(1.0);
    Player[playernum].meca.dx = toFIXED(0.0);
    Player[playernum].meca.dy = toFIXED(0.0);
    Player[playernum].bonus = 0;  	
  }
}

// Insere un joueur dans un endroit vide de la liste des joueurs
void addPlayer(int padnumber) {
  int i = 0;
  // on recherche un emplacement vide dans Player
  while ((Player[i].state != PL_STATE_DEAD) && (i<=MAX_PLAYER)) i++;
  // on y fout le mec
  if (i<=MAX_PLAYER) {
    wakeUpPlayer(i,padnumber);
  }
}

void removePlayer(int i) { // et on ne remet meme pas le numero de manette a -1
  	Player[i].state = PL_STATE_DEAD;
}

void limboPlayer(int i) {
   	 Player[i].state = PL_STATE_LIMBO;
   	 Player[i].r_limbo_time = LIMBO_TIME;
}

// Reinitialise la liste des joueurs et l'assignation des avions
void resetPlayerList() {
  int i;
  for (i = 0 ; i <= MAX_PLAYER; i++) {
    removePlayer(i);
  }
}

// returns the scale for player's plane drawing
FIXED playerGetScale(int playeridx){
  FIXED scale;
  // 5 * energy / maxplayerenergy
  scale = slDivFX(toFIXED(MAX_PLAYER_ENERGY),slMulFX(toFIXED(Player[playeridx].energy),toFIXED(5.0)));
  return scale;
}

/*
 * Get Player idx bny attribute data
 *           getPlayerByPadIdx(int padidx)  
 *           getPlayerByPlaneIdx(int planeidx)
 */

int getPlayerByPadIdx(int padidx) {
  int i;
  int noplayer = -1;
  for (i = 0 ; i <=MAX_PLAYER; i++) {
    if ((Player[i].pad_num == padidx) && (Player[i].state != PL_STATE_DEAD))
		noplayer = i;
    }
  return noplayer;
}

int getPlayerByPlaneIdx(int planeidx) {
  int i;
  int noplayer = -1;
  for (i = 0 ; i <=MAX_PLAYER; i++) {
    if ((Player[i].plane_num == planeidx) && (Player[i].state != PL_STATE_DEAD))
		noplayer = i;
    }
  return noplayer;
}

// poll les ports pour connaitre l'tat des connexions
void pollPlayerList(){
  Uint16 padData;
  int i,playernum; // i = padnumber
  
  for (i = 0; i<=MAX_SATURN_PAD;i++) {
    if (Per_Connect1 || Per_Connect2) {
    	padData = Smpc_Peripheral[i].data;
    	if (!(PER_DGT_ST & padData)) {
    	  playernum = getPlayerByPadIdx(i);
    	  if (playernum != -1){	// trouv
    	   if (!(PER_DGT_ST & (pad_GetUpFront(Player[playernum].prev_pad_data,padData)))) {		
    	     // on est bien dans un front montant.
	    	 limboPlayer(playernum);
	     }
    	  } else {
    	    	addPlayer(i);
    	  } 	  
    	}
    }
  }
}

/*********************************************************
 * DEBUG
 *********************************************************/
void printTestColl() {
  int i,j;
  i = 0;
  for (i = 0; i<=MAX_COLL;i++) {
    if (Coll[i].isUsed) {
      slPrint("X",slLocate(1,i));
    } else {
      slPrint("-",slLocate(1,i));
    }
    for (j = 0 ; j<=MAX_PLAYER ; j++) {
      if (Coll[i].player[j] >0) {
	slPrint("+",slLocate(Coll[i].player[j]+2,i));     
      }
    }  
  }
}

void printTestPlayer(int line){
  int i;
  for (i = 0 ; i <= MAX_PLAYER; i=i+2) {
    if (Player[i].state == PL_STATE_PLAY )
    {slPrint("P",slLocate(1,line+i));}
    if (Player[i].state == PL_STATE_DEAD )
    {slPrint("D",slLocate(1,line+i));}
    if (Player[i].state == PL_STATE_LIMBO )
    {slPrint("L",slLocate(1,line+i));}
 
   slPrintFX(Player[i].meca.x, slLocate(3,line+i));
    slPrintFX(Player[i].meca.y, slLocate(15,line+i));
    slPrintFX(Player[i].meca.dx, slLocate(22,line+i));
    slPrintFX(Player[i].meca.dy, slLocate(28,line+i));

  }
}


void printTestBomb(int line){
  int i;
  for (i = 0 ; i <= 0/*MAX_BOMB*/; i++) {
    if (Bomb[i].state == BB_STATE_NIU )
    {slPrint("N",slLocate(1,line+i));}
    if (Bomb[i].state == BB_STATE_CDOWN )
    {slPrint("C",slLocate(1,line+i));}
    if (Bomb[i].state == BB_STATE_EXPL )
    {slPrint("E",slLocate(1,line+i));}

    slPrintFX(Bomb[i].meca.x, slLocate(3,line+i));
    slPrintFX(Bomb[i].meca.y, slLocate(15,line+i));
    slPrintFX(toFIXED(Bomb[i].r_cdown_time), slLocate(3,line+i+1));
    slPrintFX(toFIXED(Bomb[i].r_expl_time), slLocate(15,line+i+1));

  }
}

// affiche l'tat d'un testpaddata
void printTestPadData(Uint16 paddata,int line) {
    int i;
    for (i=0;i<16;i++) {
      if (!((1<<i) & paddata)) {
        slPrint("X",slLocate(i,line));
      } else {
        slPrint(" ",slLocate(i,line));
      }
    }
}

/*********************************************************
 * SUB MAIN
 **********************************************************/
void displayBomb(int i) {
  FIXED Z;
  ANGLE rotZ;
  FIXED scale;
  Z = toFIXED(220.0); 
  rotZ = DEGtoANG(0.0);
  scale = toFIXED(1.0);
  
  slPushMatrix(); 
  {
    slTranslate(logicalXToScreenX(Bomb[i].meca.x),logicalYToScreenY(Bomb[i].meca.y),Z);
    slPushMatrix();
    {
      slRotX(DEGtoANG(0.0));
      slRotY(DEGtoANG(0.0));

      if (Bomb[i].state == BB_STATE_CDOWN) {
        if (Bomb[i].cdown_time != 0)
          rotZ = DEGtoANG(360.0*Bomb[i].r_cdown_time/Bomb[i].cdown_time);
      } 
      slPushMatrix();
      {
	if (Bomb[i].state == BB_STATE_EXPL) {
	  if (Bomb[i].expl_time != 0) {
	    rotZ = DEGtoANG(360.0*Bomb[i].r_expl_time/Bomb[i].expl_time);      
	    //          scale = toFIXED(10.0)+slMulFX(toFIXED(10.0),Bomb[i].c_expl_radius);
	    scale = toFIXED(1+(15.0*Bomb[i].r_expl_time/Bomb[i].expl_time));
	  }
	}
	slScale(scale,scale,toFIXED(1.0));
      }
      slPopMatrix();
      slRotZ(rotZ);
      slPutPolygon(Bomb[i].ppdata);

    }
    slPopMatrix();
  }
  slPopMatrix();
}

// affichage de l'avion du joueur
void displayPlayer (int i){
  FIXED Z;
  FIXED scale;

  Z = toFIXED(220.0); 
  scale = toFIXED(3.0);
  

  slPushMatrix(); 
  {
    Z = /*(PER_DGT_TA & padData))?toFIXED(180.0):*/toFIXED(220.0);
    slTranslate(logicalXToScreenX(Player[i].meca.x),logicalYToScreenY(Player[i].meca.y),Z);
    slRotX(DEGtoANG(0.0));
    slRotY(DEGtoANG(0.0));
    slRotZ(DEGtoANG(0.0));
    slPushMatrix();
    {
      slScale(playerGetScale(i),playerGetScale(i),1.0);
      //slScale(scale,1.0,1.0);
      slPutPolygon(Plane[Player[i].plane_num].ppdata);
    }
    slPopMatrix();
  }
  slPopMatrix();
}



void bombStateMachine(int i) {
  if (Bomb[i].state != BB_STATE_NIU) {
    
    if (Bomb[i].state == BB_STATE_EXPL) { 		// explosion
      if (Bomb[i].r_expl_time <= 0) {
        Bomb[i].state = BB_STATE_NIU;
      } else {
        Bomb[i].r_expl_time -=1;
        Bomb[i].c_expl_radius 
        	= slDivFX(toFIXED(Bomb[i].expl_time),(Bomb[i].expl_radius*toFIXED(Bomb[i].r_expl_time)));
      }
      displayBomb(i);
    }
    
    if (Bomb[i].state == BB_STATE_CDOWN) {		// compte  la reboure
      if (Bomb[i].r_cdown_time == 0){
        Bomb[i].state = BB_STATE_EXPL;
      } else {
        Bomb[i].r_cdown_time -= 1;
        displayBomb(i);
      }
    }
  }
}


void playerStateMachine(int i) {
  int j;
  Uint16 padData = 0;
  int traverse = FALSE;

  if (Player[i].state != PL_STATE_DEAD) {
    
    // ** dans les limbes, plus de controle
    if (Player[i].state == PL_STATE_LIMBO) {
      if (Player[i].r_limbo_time <= 0) {
        removePlayer(i);
      } else {
        Player[i].r_limbo_time --;
      }
    }
        
    // ** vivant : on controle.
    if (Player[i].state == PL_STATE_PLAY) {
      // collisions
      for (j = 0; j<=MAX_BOMB; j++) {
        if (Bomb[j].state == BB_STATE_EXPL){
          if (getDistance(Player[i].meca.x,Player[i].meca.y,Bomb[j].meca.x,Bomb[j].meca.y) < toFIXED(0.2)) {
            Player[i].energy -= 1;
          }
        }
        if ((Player[i].energy <= 0)){
          limboPlayer(i);
        }	
      }           
      
      // modification de la vitesse par l'acceleration, en fonction du pad.
      padData = Smpc_Peripheral[Player[i].pad_num].data;
      
      applyPadAccel(&Player[i].meca.dx,&Player[i].meca.dy,
		 Plane[Player[i].plane_num].accel,Plane[Player[i].plane_num].speedLimit,
		 ((!(PER_DGT_KR & padData))?1:((!(PER_DGT_KL & padData))?-1:0)),
		 ((!(PER_DGT_KD & padData))?1:((!(PER_DGT_KU & padData))?-1:0)) );
	
      // laching de bombe
      if (!(PER_DGT_TC & pad_GetUpFront(Player[i].prev_pad_data,padData))) {
	askLaunchBomb(Player[i].meca.x,Player[i].meca.y);
      }
      
    } // if state == PL_STATE_PLAY
    
    // la vitesse modifie la position
    applySpeed(&(Player[i].meca.x),Player[i].meca.dx);
    applySpeed(&(Player[i].meca.y),Player[i].meca.dy);
    
    padData = Smpc_Peripheral[Player[i].pad_num].data;
    traverse = !(PER_DGT_TR & padData);

    // on bloque contre les bords ou traverse
    if (Player[i].meca.x >  S_LIM_R) {	
      Player[i].meca.x =  traverse?S_LIM_L:S_LIM_R; 	  		
      Player[i].meca.dx =  traverse?Player[i].meca.dx:-1.0*Player[i].meca.dx;
    }
    if (Player[i].meca.x < S_LIM_L) {
      Player[i].meca.x = traverse? S_LIM_R:S_LIM_L;
      Player[i].meca.dx =  traverse?Player[i].meca.dx :-1.0*Player[i].meca.dx;
    }
    if (Player[i].meca.y >  S_LIM_B) {
      Player[i].meca.y =  traverse? S_LIM_T:S_LIM_B;
      Player[i].meca.dy =  traverse?Player[i].meca.dy:-1.0*Player[i].meca.dy;
    }
    if (Player[i].meca.y < S_LIM_T) {
      Player[i].meca.y = traverse? S_LIM_B:S_LIM_T;
      Player[i].meca.dy =  traverse?Player[i].meca.dy:-1.0*Player[i].meca.dy;
    }
    
    displayPlayer(i);
    Player[i].prev_pad_data = padData;  
  }      
}
/***********************************************************
 * COLLISIONS
 **********************************************************/

// vide le tableau des collisions
void emptyCollTable() {
  int i,j,k/*,l*/;
  for (i=0;i<=MAX_COLL;i++) {
    Coll[i].isUsed = FALSE;
    for (j=0;j<=MAX_PLAYER;j++) { Coll[i].player[j] = -1;}
    for (k=0;k<=MAX_BOMB;k++)   { Coll[i].bomb[k] = -1;}
    // + murs
  }
}

// collision entre p1 et p2 ?
int isCollPlayers (int p1, int p2) {
  int isColl = -1;  
  FIXED playersize = toFIXED(5.0); // size of the plane's bounding box.
  FIXED scale1;
  FIXED scale2;
  scale1 = playerGetScale(p1);
  scale2 = playerGetScale(p2);
    
  isColl = -1;
  isColl = slMulFX(L_XCOEF,getDistance(Player[p1].meca.x,Player[p1].meca.y,Player[p2].meca.x,Player[p2].meca.y)) 
    < (slMulFX(playersize,scale1)+slMulFX(playersize,scale2));
  return isColl;
}

// ajout d'une collision dans le tableau des collisions.
void addCollPlayer (int p1, int p2) {
  int i,j;
  int found;
  int ptoadd;
  // on recherche dans le tableau 
  // - la coll existante mettant en jeu p1 et/ou p2
  // - sinon, le premier endroit libre pour la mettre
  i = 0;
  j = 0; // pour faire plaisir au compilo

  found = (0 != 0);
  
  while ((Coll[i].isUsed == TRUE) && (!found) && (i<=MAX_COLL)) {
    j = 0;
    while ((Coll[i].player[j] != -1) && (!found) && (j<=MAX_PLAYER)) {
      found = ((Coll[i].player[j] == p1) || (Coll[i].player[j] == p2));
      j++;
    }
    i++;
  }
  
  if (found){      // une collision met en scene p1 ou p2, on ajoute dans la case Coll[i-1]
    j--;i--;
    if (Coll[i].player[j] == p1) {ptoadd = p2;} else {ptoadd = p1;}
    while ((j != -1) && (j <=MAX_PLAYER)) {j++;} // on va au bout du tableau
    if (j <= MAX_PLAYER) 
      Coll[i].player[j] = ptoadd;
  } else {         // pas de collision avec p1 ou p2 existante, on ajoute dans la case Coll[i]
    if (!(Coll[i].isUsed)){// si le tableau est plein, on abandonne
      Coll[i].isUsed = TRUE;
      Coll[i].player[0] = p1;
      Coll[i].player[1] = p2;
    }
  }
}

// retourne le nombre de collisions dans le tableau coll
// remplit le tableau avec les coll en cours.
void setCollTable () {
  int i,j;
  for (i=0;i<=MAX_PLAYER;i++) {
    if (Player[i].state != PL_STATE_DEAD) {
      for (j=i+1; j<=MAX_PLAYER;j++) {
	if (Player[j].state != PL_STATE_DEAD) {
	  if (isCollPlayers(i,j)) {
	    addCollPlayer(i,j);
	  }
	}
      }
    }
  }
}

// a partir du tableau collisions, on dtermine les nouveaux vecteurs mcaniques
void recomputeMecaVectorsColl (){
  int i,j;
  i = 0;
  while ((Coll[i].isUsed)&&(i<=MAX_COLL)) {
    // pour chaque collision
    j = 0;

   
    /// ahhhh !
    if (GLOBAL_SOUND_STATE == gss_on) {
      playasound();
    } else {
      scrollTxtAddTxtLine("bing");
    }

    while ((Coll[i].player[j] != -1) && (j<=MAX_PLAYER)) {
      // pour chaque element de la collision.
      //      scrollTxtAddFIXEDLine(toFIXED(j));
      Player[Coll[i].player[j]].meca.dx =  slMulFX(Player[Coll[i].player[j]].meca.dx,toFIXED(-1.0));
      Player[Coll[i].player[j]].meca.dy =  slMulFX(Player[Coll[i].player[j]].meca.dy,toFIXED(-1.0));
      j++;
    }
    i++;
  }
}

void computeCollusions() {
  emptyCollTable();
  setCollTable();
  recomputeMecaVectorsColl();
}

/********************************************************
 * INFOS
 ********************************************************/

void dsplyLicense() {
  scrollTxtAddTxtLine("~Le Pin~                  ");
  scrollTxtAddTxtLine("a multiplayer game        ");
  scrollTxtAddTxtLine("rel. under the GNU/GPL    ");
  scrollTxtAddTxtLine("[2003]                    ");
}

/*******************************************************
 * GLOBAL SCREEN TRANSFORMS
 *********************************************************/

// quick hack for fun
void  globalTransforms(){
  int i;
  Uint16 padData = 0;

  static ANGLE arotZ=0x0;
  static ANGLE drotZ=0x0000;

  static FIXED aposZ=0x0;   
  static FIXED dposZ=0x0000;    

  static FIXED cam_s[] = {toFIXED(0.0),toFIXED(0.0),toFIXED(0.0)};
  static FIXED cam_d[] = {toFIXED(0.0),toFIXED(0.0),toFIXED(100.0)};
  static ANGLE cam_a = DEGtoANG(0.0);

  slLookAt(cam_s, cam_d, cam_a);

  for (i=0;i<=MAX_PLAYER;i++) {
    if (Player[i].state != PL_STATE_DEAD) {
      
      padData = Smpc_Peripheral[Player[i].pad_num].data;     
      drotZ += (!(PER_DGT_TX & padData))?0x0010:0x0;
      drotZ -= (!(PER_DGT_TA & padData))?0x0010:0x0;
      dposZ += (!(PER_DGT_TY & padData))?toFIXED(2.0):0x0;
      dposZ -= (!(PER_DGT_TB & padData))?toFIXED(2.0):0x0;
      
      if ((!(PER_DGT_TX & padData)) && (!(PER_DGT_TA & padData))){
	dposZ = 0x0;
	aposZ = 0x0;
      }
      if ((!(PER_DGT_TY & padData)) && (!(PER_DGT_TB & padData))){
	drotZ = 0x0;
	arotZ = 0x0;
      }      
    }
  }

  if (aposZ>toFIXED(1000.0)){
    aposZ = toFIXED(1000.0);
    dposZ = slMulFX(toFIXED(-1.0),slDivFX(toFIXED(2.0),dposZ));
  }

  if (aposZ<toFIXED(-100.0)){  
    aposZ = toFIXED(-100.0);
    dposZ = slMulFX(toFIXED(-1.0),slDivFX(toFIXED(2.0),dposZ));
  }

  arotZ += drotZ;
  aposZ += dposZ;

  slRotZ(arotZ);
  slTranslate(toFIXED(0.0), toFIXED(0.0), aposZ);
  /*
  slPrintFX(arotZ, slLocate(1,15));
  slPrintFX(drotZ, slLocate(10,15));
  slPrintFX(aposZ, slLocate(1,16));
  slPrintFX(dposZ, slLocate(10,16));
  */
}


/*********************************************************
 * MAIN
 *********************************************************/

void ss_main() {

  int i;
  Uint16 padData = 0;
  Uint16 oldPadData = 0;


  slInitSystem(TV_320x224,NULL,1);
  init_sound();

  buildPlaneArray();
  buildBombArray();
  resetPlayerList();
  scrollTxtInit();
  
  dsplyLicense();


  while (-1) {
    slPushUnitMatrix();
    {

      globalTransforms();   

      padData = Smpc_Peripheral[0].data;     
      
      if (!(PER_DGT_TZ & (pad_GetUpFront(oldPadData,padData)))) {		
	if (GLOBAL_SOUND_STATE == gss_off) {
	  scrollTxtAddTxtLine("(*)                       ");
	  GLOBAL_SOUND_STATE = gss_on;
	} else {
	  scrollTxtAddTxtLine("( )                       ");
	  GLOBAL_SOUND_STATE = gss_off;
	}
      }
      oldPadData = padData;

      // bombs
      for (i=0;i<=MAX_BOMB;i++) {
	bombStateMachine(i);
      } 
      
      // player
      pollPlayerList();  
      for (i=0;i<=MAX_PLAYER;i++) {
	playerStateMachine(i);
      }    
    
      computeCollusions();
    }
    // prntTestPlayer(0);
    // printTestBomb(15);
    // printTestColl();
    
    slPopMatrix();
    
    scrollTxtSchedule();
    scrollTxtDisplay();
    slSynch();
  }
}



